@@ -1,20 +1,24 @@
+import bisect
 import logging
 import re
 import typing
 from abc import ABC, abstractmethod
-from collections import Counter, defaultdict
+from collections import Counter, defaultdict, namedtuple
 from functools import lru_cache
 from operator import itemgetter
 from pathlib import Path
-from typing import Dict, List, Optional, Union, cast
+from typing import Dict, Iterable, List, Optional, Union, cast
 
 import torch
 from deprecated import deprecated
-from torch.utils.data import Dataset
+from torch.utils.data import Dataset, IterableDataset
 from torch.utils.data.dataset import ConcatDataset, Subset
 
 import flair
 from flair.file_utils import Tqdm
+from flair.tokenization import SegtokTokenizer, SpaceTokenizer, Tokenizer
+
+T_co = typing.TypeVar("T_co", covariant=True)
 
 log = logging.getLogger("flair")
 
@@ -36,6 +40,9 @@ def _len_dataset(dataset: Optional[Dataset]) -> int:
     return len(loader)
 
 
+BoundingBox = namedtuple("BoundingBox", ["left", "top", "right", "bottom"])
+
+
 class Dictionary:
     """
     This class holds a dictionary that maps strings to IDs, used to generate one-hot encodings of strings.
@@ -53,7 +60,6 @@ class Dictionary:
             self.add_item("<unk>")
 
     def remove_item(self, item: str):
-
         bytes_item = item.encode("utf-8")
         if bytes_item in self.item2idx:
             self.idx2item.remove(bytes_item)
@@ -198,35 +204,24 @@ class Label:
     score needs to be between 0.0 and 1.0. Default value for the score is 1.0.
     """
 
-    def __init__(self, data_point, value: Optional[str], score: float = 1.0):
+    def __init__(self, data_point: "DataPoint", value: str, score: float = 1.0):
         self._value = value
         self._score = score
         self.data_point: DataPoint = data_point
         super().__init__()
 
     def set_value(self, value: str, score: float = 1.0):
-        self.value = value
-        self.score = score
+        self._value = value
+        self._score = score
 
     @property
-    def value(self):
+    def value(self) -> str:
         return self._value
 
-    @value.setter
-    def value(self, value):
-        if not value and value != "":
-            raise ValueError("Incorrect label value provided. Label value needs to be set.")
-        else:
-            self._value = value
-
     @property
-    def score(self):
+    def score(self) -> float:
         return self._score
 
-    @score.setter
-    def score(self, score):
-        self._score = score
-
     def to_dict(self):
         return {"value": self.value, "confidence": self.score}
 
@@ -269,6 +264,7 @@ class DataPoint:
     def __init__(self):
         self.annotation_layers = {}
         self._embeddings: Dict[str, torch.Tensor] = {}
+        self._metadata: Dict[str, typing.Any] = {}
 
     @property
     @abstractmethod
@@ -279,12 +275,19 @@ class DataPoint:
         self._embeddings[name] = vector
 
     def get_embedding(self, names: Optional[List[str]] = None) -> torch.Tensor:
-        embeddings = self.get_each_embedding(names)
+        # if one embedding name, directly return it
+        if names and len(names) == 1:
+            if names[0] in self._embeddings:
+                return self._embeddings[names[0]].to(flair.device)
+            else:
+                return torch.tensor([], device=flair.device)
 
+        # if multiple embedding names, concatenate them
+        embeddings = self.get_each_embedding(names)
         if embeddings:
             return torch.cat(embeddings, dim=0)
-
-        return torch.tensor([], device=flair.device)
+        else:
+            return torch.tensor([], device=flair.device)
 
     def get_each_embedding(self, embedding_names: Optional[List[str]] = None) -> List[torch.Tensor]:
         embeddings = []
@@ -317,8 +320,16 @@ class DataPoint:
         else:
             return False
 
-    def add_label(self, typename: str, value: str, score: float = 1.0):
+    def add_metadata(self, key: str, value: typing.Any) -> None:
+        self._metadata[key] = value
 
+    def get_metadata(self, key: str) -> typing.Any:
+        return self._metadata[key]
+
+    def has_metadata(self, key: str) -> bool:
+        return key in self._metadata
+
+    def add_label(self, typename: str, value: str, score: float = 1.0):
         if typename not in self.annotation_layers:
             self.annotation_layers[typename] = [Label(self, value, score)]
         else:
@@ -358,7 +369,6 @@ class DataPoint:
         raise NotImplementedError
 
     def _printout_labels(self, main_label=None, add_score: bool = True):
-
         all_labels = []
         keys = [main_label] if main_label is not None else self.annotation_layers.keys()
         if add_score:
@@ -418,6 +428,9 @@ class DataPoint:
     def __hash__(self):
         return hash(self.unlabeled_identifier)
 
+    def __len__(self):
+        raise NotImplementedError
+
 
 DT = typing.TypeVar("DT", bound=DataPoint)
 DT2 = typing.TypeVar("DT2", bound=DataPoint)
@@ -436,20 +449,19 @@ class _PartOfSentence(DataPoint, ABC):
 
     def add_label(self, typename: str, value: str, score: float = 1.0):
         super().add_label(typename, value, score)
-
-        label = Label(self, value, score)
-        if typename not in self.sentence.annotation_layers:
-            self.sentence.annotation_layers[typename] = [label]
-        else:
-            self.sentence.annotation_layers[typename].append(label)
+        self.sentence.annotation_layers.setdefault(typename, []).append(Label(self, value, score))
 
     def set_label(self, typename: str, value: str, score: float = 1.0):
+        if len(self.annotation_layers.get(typename, [])) > 0:
+            # First we remove any existing labels for this PartOfSentence in self.sentence
+            self.sentence.annotation_layers[typename] = [
+                label for label in self.sentence.annotation_layers.get(typename, []) if label.data_point != self
+            ]
+        self.sentence.annotation_layers.setdefault(typename, []).append(Label(self, value, score))
         super().set_label(typename, value, score)
-        self.sentence.annotation_layers[typename] = [Label(self, value, score)]
         return self
 
     def remove_labels(self, typename: str):
-
         # labels also need to be deleted at Sentence object
         for label in self.get_labels(typename):
             self.sentence.annotation_layers[typename].remove(label)
@@ -468,7 +480,7 @@ class Token(_PartOfSentence):
         self,
         text: str,
         head_id: int = None,
-        whitespace_after: bool = True,
+        whitespace_after: int = 1,
         start_position: int = 0,
         sentence=None,
     ):
@@ -477,10 +489,9 @@ class Token(_PartOfSentence):
         self.form: str = text
         self._internal_index: Optional[int] = None
         self.head_id: Optional[int] = head_id
-        self.whitespace_after: bool = whitespace_after
+        self.whitespace_after: int = whitespace_after
 
-        self.start_pos = start_position
-        self.end_pos = start_position + len(text)
+        self._start_position = start_position
 
         self._embeddings: Dict = {}
         self.tags_proba_dist: Dict[str, List[Label]] = {}
@@ -490,15 +501,15 @@ class Token(_PartOfSentence):
         if isinstance(self._internal_index, int):
             return self._internal_index
         else:
-            raise ValueError
+            return -1
 
     @property
-    def text(self):
+    def text(self) -> str:
         return self.form
 
     @property
     def unlabeled_identifier(self) -> str:
-        return f'Token[{self.idx-1}]: "{self.text}"'
+        return f'Token[{self.idx - 1}]: "{self.text}"'
 
     def add_tags_proba_dist(self, tag_type: str, tags: List[Label]):
         self.tags_proba_dist[tag_type] = tags
@@ -513,16 +524,23 @@ class Token(_PartOfSentence):
 
     @property
     def start_position(self) -> int:
-        return self.start_pos
+        return self._start_position
+
+    @start_position.setter
+    def start_position(self, value: int) -> None:
+        self._start_position = value
 
     @property
     def end_position(self) -> int:
-        return self.end_pos
+        return self.start_position + len(self.text)
 
     @property
     def embedding(self):
         return self.get_embedding()
 
+    def __len__(self):
+        return 1
+
     def __repr__(self):
         return self.__str__()
 
@@ -567,11 +585,11 @@ class Span(_PartOfSentence):
 
     @property
     def text(self) -> str:
-        return " ".join([t.text for t in self.tokens])
+        return "".join([t.text + t.whitespace_after * " " for t in self.tokens]).strip()
 
     @property
     def unlabeled_identifier(self) -> str:
-        return f'Span[{self.tokens[0].idx -1}:{self.tokens[-1].idx}]: "{self.text}"'
+        return f'Span[{self.tokens[0].idx - 1}:{self.tokens[-1].idx}]: "{self.text}"'
 
     def __repr__(self):
         return self.__str__()
@@ -587,7 +605,7 @@ class Span(_PartOfSentence):
 
     @property
     def embedding(self):
-        pass
+        return self.get_embedding()
 
 
 class Relation(_PartOfSentence):
@@ -612,8 +630,8 @@ class Relation(_PartOfSentence):
     def unlabeled_identifier(self) -> str:
         return (
             f"Relation"
-            f"[{self.first.tokens[0].idx-1}:{self.first.tokens[-1].idx}]"
-            f"[{self.second.tokens[0].idx-1}:{self.second.tokens[-1].idx}]"
+            f"[{self.first.tokens[0].idx - 1}:{self.first.tokens[-1].idx}]"
+            f"[{self.second.tokens[0].idx - 1}:{self.second.tokens[-1].idx}]"
             f': "{self.text}"'
         )
 
@@ -630,25 +648,6 @@ class Relation(_PartOfSentence):
         pass
 
 
-class Tokenizer(ABC):
-    r"""An abstract class representing a :class:`Tokenizer`.
-
-    Tokenizers are used to represent algorithms and models to split plain text into
-    individual tokens / words. All subclasses should overwrite :meth:`tokenize`, which
-    splits the given plain text into tokens. Moreover, subclasses may overwrite
-    :meth:`name`, returning a unique identifier representing the tokenizer's
-    configuration.
-    """
-
-    @abstractmethod
-    def tokenize(self, text: str) -> List[str]:
-        raise NotImplementedError()
-
-    @property
-    def name(self) -> str:
-        return self.__class__.__name__
-
-
 class Sentence(DataPoint):
     """
     A Sentence is a list of tokens and is used to represent a sentence or text fragment.
@@ -656,7 +655,7 @@ class Sentence(DataPoint):
 
     def __init__(
         self,
-        text: Union[str, List[str]],
+        text: Union[str, List[str], List[Token]],
         use_tokenizer: Union[bool, Tokenizer] = True,
         language_code: str = None,
         start_position: int = 0,
@@ -681,59 +680,18 @@ class Sentence(DataPoint):
 
         self.language_code: Optional[str] = language_code
 
-        self.start_pos = start_position
-        self.end_pos = start_position + len(text)
+        self._start_position = start_position
 
         # the tokenizer used for this sentence
         if isinstance(use_tokenizer, Tokenizer):
             tokenizer = use_tokenizer
 
         elif type(use_tokenizer) == bool:
-            from flair.tokenization import SegtokTokenizer, SpaceTokenizer
-
             tokenizer = SegtokTokenizer() if use_tokenizer else SpaceTokenizer()
 
         else:
             raise AssertionError("Unexpected type of parameter 'use_tokenizer'. Parameter should be bool or Tokenizer")
 
-        # if text is passed, instantiate sentence with tokens (words)
-        if isinstance(text, str):
-            text = Sentence._handle_problem_characters(text)
-            words = tokenizer.tokenize(text)
-        else:
-            words = text
-
-        # determine token positions and whitespace_after flag
-        current_offset = 0
-        previous_word_offset = -1
-        previous_token = None
-        for word in words:
-            try:
-                word_offset = text.index(word, current_offset)
-                start_position = word_offset
-            except ValueError:
-                word_offset = previous_word_offset + 1
-                start_position = current_offset + 1 if current_offset > 0 else current_offset
-
-            if word:
-                token = Token(text=word, start_position=start_position, whitespace_after=True)
-                self.add_token(token)
-
-            if (previous_token is not None) and word_offset - 1 == previous_word_offset:
-                previous_token.whitespace_after = False
-
-            current_offset = word_offset + len(word)
-            previous_word_offset = current_offset - 1
-            previous_token = token
-
-        # the last token has no whitespace after
-        if len(self) > 0:
-            self.tokens[-1].whitespace_after = False
-
-        # log a warning if the dataset is empty
-        if text == "":
-            log.warning("Warning: An empty Sentence was created! Are there empty strings in your dataset?")
-
         self.tokenized: Optional[str] = None
 
         # some sentences represent a document boundary (but most do not)
@@ -741,12 +699,50 @@ class Sentence(DataPoint):
 
         # internal variables to denote position inside dataset
         self._previous_sentence: Optional[Sentence] = None
+        self._has_context: bool = False
         self._next_sentence: Optional[Sentence] = None
         self._position_in_dataset: Optional[typing.Tuple[Dataset, int]] = None
 
+        # if text is passed, instantiate sentence with tokens (words)
+        if isinstance(text, str):
+            text = Sentence._handle_problem_characters(text)
+            words = tokenizer.tokenize(text)
+        elif text and isinstance(text[0], Token):
+            for t in text:
+                self._add_token(t)
+            self.tokens[-1].whitespace_after = 0
+            return
+        else:
+            words = cast(List[str], text)
+            text = " ".join(words)
+
+        # determine token positions and whitespace_after flag
+        current_offset: int = 0
+        previous_token: Optional[Token] = None
+        for word in words:
+            word_start_position: int = text.index(word, current_offset)
+            delta_offset: int = word_start_position - current_offset
+
+            token: Token = Token(text=word, start_position=word_start_position)
+            self._add_token(token)
+
+            if previous_token is not None:
+                previous_token.whitespace_after = delta_offset
+
+            current_offset = token.end_position
+            previous_token = token
+
+        # the last token has no whitespace after
+        if len(self) > 0:
+            self.tokens[-1].whitespace_after = 0
+
+        # log a warning if the dataset is empty
+        if text == "":
+            log.warning("Warning: An empty Sentence was created! Are there empty strings in your dataset?")
+
     @property
     def unlabeled_identifier(self):
-        return f'Sentence: "{self.to_tokenized_string()}"'
+        return f'Sentence[{len(self)}]: "{self.text}"'
 
     def get_relations(self, type: str) -> List[Relation]:
         relations: List[Relation] = []
@@ -768,8 +764,7 @@ class Sentence(DataPoint):
                 return token
         return None
 
-    def add_token(self, token: Union[Token, str]):
-
+    def _add_token(self, token: Union[Token, str]):
         if isinstance(token, Token):
             assert token.sentence is None
 
@@ -785,10 +780,7 @@ class Sentence(DataPoint):
         token.sentence = self
         token._internal_index = len(self.tokens) + 1
         if token.start_position == 0 and len(self) > 0:
-            token.start_pos = (
-                len(self.to_original_text()) + 1 if self[-1].whitespace_after else len(self.to_original_text())
-            )
-            token.end_pos = token.start_pos + len(token.text)
+            token.start_position = len(self.to_original_text()) + self[-1].whitespace_after
 
         # append token to sentence
         self.tokens.append(token)
@@ -806,7 +798,6 @@ class Sentence(DataPoint):
         return self.get_embedding()
 
     def to(self, device: str, pin_memory: bool = False):
-
         # move sentence embeddings to device
         super().to(device=device, pin_memory=pin_memory)
 
@@ -815,7 +806,6 @@ class Sentence(DataPoint):
             token.to(device, pin_memory)
 
     def clear_embeddings(self, embedding_names: List[str] = None):
-
         super().clear_embeddings(embedding_names)
 
         # clear token embeddings
@@ -823,10 +813,10 @@ class Sentence(DataPoint):
             token.clear_embeddings(embedding_names)
 
     @lru_cache(maxsize=1)  # cache last context, as training repeats calls
-    def left_context(self, context_length: int, respect_document_boundaries: bool = True):
+    def left_context(self, context_length: int, respect_document_boundaries: bool = True) -> List[Token]:
         sentence = self
-        left_context: List[str] = []
-        while True:
+        left_context: List[Token] = []
+        while len(left_context) < context_length:
             sentence = sentence.previous_sentence()
             if sentence is None:
                 break
@@ -834,34 +824,27 @@ class Sentence(DataPoint):
             if respect_document_boundaries and sentence.is_document_boundary:
                 break
 
-            left_context = [t.text for t in sentence.tokens] + left_context
-            if len(left_context) > context_length:
-                left_context = left_context[-context_length:]
-                break
-        return left_context
+            left_context = sentence.tokens + left_context
+        return left_context[-context_length:]
 
     @lru_cache(maxsize=1)  # cache last context, as training repeats calls
-    def right_context(self, context_length: int, respect_document_boundaries: bool = True):
+    def right_context(self, context_length: int, respect_document_boundaries: bool = True) -> List[Token]:
         sentence = self
-        right_context: List[str] = []
-        while True:
+        right_context: List[Token] = []
+        while len(right_context) < context_length:
             sentence = sentence.next_sentence()
             if sentence is None:
                 break
             if respect_document_boundaries and sentence.is_document_boundary:
                 break
 
-            right_context += [t.text for t in sentence.tokens]
-            if len(right_context) > context_length:
-                right_context = right_context[:context_length]
-                break
-        return right_context
+            right_context += sentence.tokens
+        return right_context[:context_length]
 
     def __str__(self):
         return self.to_tagged_string()
 
     def to_tagged_string(self, main_label=None) -> str:
-
         already_printed = [self]
 
         output = super().__str__()
@@ -882,10 +865,9 @@ class Sentence(DataPoint):
 
     @property
     def text(self):
-        return self.to_tokenized_string()
+        return self.to_original_text()
 
     def to_tokenized_string(self) -> str:
-
         if self.tokenized is None:
             self.tokenized = " ".join([t.text for t in self.tokens])
 
@@ -895,8 +877,8 @@ class Sentence(DataPoint):
         plain = ""
         for token in self.tokens:
             plain += token.text
-            if token.whitespace_after:
-                plain += " "
+            if token.whitespace_after > 0:
+                plain += token.whitespace_after * " "
         return plain.rstrip()
 
     def infer_space_after(self):
@@ -913,36 +895,29 @@ class Sentence(DataPoint):
             if token.text == '"':
                 quote_count += 1
                 if quote_count % 2 != 0:
-                    token.whitespace_after = False
+                    token.whitespace_after = 0
                 elif last_token is not None:
-                    last_token.whitespace_after = False
+                    last_token.whitespace_after = 0
 
             if last_token is not None:
-
                 if token.text in [".", ":", ",", ";", ")", "n't", "!", "?"]:
-                    last_token.whitespace_after = False
+                    last_token.whitespace_after = 0
 
                 if token.text.startswith("'"):
-                    last_token.whitespace_after = False
+                    last_token.whitespace_after = 0
 
             if token.text in ["("]:
-                token.whitespace_after = False
+                token.whitespace_after = 0
 
             last_token = token
         return self
 
     def to_original_text(self) -> str:
-        str = ""
-        pos = 0
-        for t in self.tokens:
-            while t.start_pos > pos:
-                str += " "
-                pos += 1
-
-            str += t.text
-            pos += len(t.text)
-
-        return str
+        # if sentence has no tokens, return empty string
+        if len(self) == 0:
+            return ""
+        # otherwise, return concatenation of tokens with the correct offsets
+        return self[0].start_position * " " + "".join([t.text + t.whitespace_after * " " for t in self.tokens]).strip()
 
     def to_dict(self, tag_type: str = None):
         labels = []
@@ -983,27 +958,19 @@ class Sentence(DataPoint):
     def __repr__(self):
         return self.__str__()
 
-    def __copy__(self):
-        s = Sentence()
-        for token in self.tokens:
-            nt = Token(token.text)
-            for tag_type in token.tags:
-                nt.add_label(
-                    tag_type,
-                    token.get_tag(tag_type).value,
-                    token.get_tag(tag_type).score,
-                )
-
-            s.add_token(nt)
-        return s
-
     @property
     def start_position(self) -> int:
-        return 0
+        return self._start_position
+
+    @start_position.setter
+    def start_position(self, value: int) -> None:
+        self._start_position = value
 
     @property
     def end_position(self) -> int:
-        return len(self.to_original_text())
+        # The sentence's start position is not propagated to its tokens.
+        # Therefore, we need to add the sentence's start position to its last token's end position, including whitespaces.
+        return self.start_position + self[-1].end_position + self[-1].whitespace_after
 
     def get_language_code(self) -> str:
         if self.language_code is None:
@@ -1078,10 +1045,32 @@ class Sentence(DataPoint):
         Return True or False depending on whether context is set (for instance in dataloader or elsewhere)
         :return: True if context is set, else False
         """
-        return self._previous_sentence is not None or self._position_in_dataset is not None
+        return (
+            self._has_context
+            or self._previous_sentence is not None
+            or self._next_sentence is not None
+            or self._position_in_dataset is not None
+        )
+
+    def copy_context_from_sentence(self, sentence: "Sentence") -> None:
+        self._previous_sentence = sentence._previous_sentence
+        self._next_sentence = sentence._next_sentence
+        self._position_in_dataset = sentence._position_in_dataset
+
+    @classmethod
+    def set_context_for_sentences(cls, sentences: List["Sentence"]) -> None:
+        previous_sentence = None
+        for sentence in sentences:
+            if sentence.is_context_set():
+                continue
+            sentence._previous_sentence = previous_sentence
+            sentence._next_sentence = None
+            sentence._has_context = True
+            if previous_sentence is not None:
+                previous_sentence._next_sentence = sentence
+            previous_sentence = sentence
 
     def get_labels(self, label_type: str = None):
-
         # if no label if specified, return all labels
         if label_type is None:
             return sorted(self.labels)
@@ -1094,7 +1083,6 @@ class Sentence(DataPoint):
         return []
 
     def remove_labels(self, typename: str):
-
         # labels also need to be deleted at all tokens
         for token in self:
             token.remove_labels(typename)
@@ -1186,18 +1174,12 @@ class Image(DataPoint):
         pass
 
 
-class FlairDataset(Dataset):
-    @abstractmethod
-    def is_in_memory(self) -> bool:
-        pass
-
-
-class Corpus:
+class Corpus(typing.Generic[T_co]):
     def __init__(
         self,
-        train: Dataset = None,
-        dev: Dataset = None,
-        test: Dataset = None,
+        train: Optional[Dataset[T_co]] = None,
+        dev: Optional[Dataset[T_co]] = None,
+        test: Optional[Dataset[T_co]] = None,
         name: str = "corpus",
         sample_missing_splits: Union[bool, str] = True,
     ):
@@ -1221,20 +1203,20 @@ class Corpus:
             dev, train = randomly_split_into_two_datasets(train, dev_size)
 
         # set train dev and test data
-        self._train: Optional[Dataset] = train
-        self._test: Optional[Dataset] = test
-        self._dev: Optional[Dataset] = dev
+        self._train: Optional[Dataset[T_co]] = train
+        self._test: Optional[Dataset[T_co]] = test
+        self._dev: Optional[Dataset[T_co]] = dev
 
     @property
-    def train(self) -> Optional[Dataset]:
+    def train(self) -> Optional[Dataset[T_co]]:
         return self._train
 
     @property
-    def dev(self) -> Optional[Dataset]:
+    def dev(self) -> Optional[Dataset[T_co]]:
         return self._dev
 
     @property
-    def test(self) -> Optional[Dataset]:
+    def test(self) -> Optional[Dataset[T_co]]:
         return self._test
 
     def downsample(
@@ -1244,7 +1226,6 @@ class Corpus:
         downsample_dev=True,
         downsample_test=True,
     ):
-
         if downsample_train and self._train is not None:
             self._train = self._downsample_to_proportion(self._train, percentage)
 
@@ -1278,7 +1259,6 @@ class Corpus:
 
     @staticmethod
     def _filter_long_sentences(dataset, max_charlength: int) -> Dataset:
-
         # find out empty sentence indices
         empty_sentence_indices = []
         non_empty_sentence_indices = []
@@ -1296,7 +1276,6 @@ class Corpus:
 
     @staticmethod
     def _filter_empty_sentences(dataset) -> Dataset:
-
         # find out empty sentence indices
         empty_sentence_indices = []
         non_empty_sentence_indices = []
@@ -1349,7 +1328,6 @@ class Corpus:
 
     @staticmethod
     def _downsample_to_proportion(dataset: Dataset, proportion: float):
-
         sampled_size: int = round(_len_dataset(dataset) * proportion)
         splits = randomly_split_into_two_datasets(dataset, sampled_size)
         return splits[0]
@@ -1418,7 +1396,7 @@ class Corpus:
         for sent in sentences:
             for token in sent.tokens:
                 if label_type in token.annotation_layers.keys():
-                    label = token.get_tag(label_type)
+                    label = token.get_label(label_type)
                     label_count[label.value] += 1
         return label_count
 
@@ -1429,12 +1407,15 @@ class Corpus:
             _len_dataset(self.test) if self.test else 0,
         )
 
-    def make_label_dictionary(self, label_type: str, min_count: int = -1) -> Dictionary:
+    def make_label_dictionary(self, label_type: str, min_count: int = -1, add_unk: bool = True) -> Dictionary:
         """
         Creates a dictionary of all labels assigned to the sentences in the corpus.
         :return: dictionary of labels
         """
-        label_dictionary: Dictionary = Dictionary(add_unk=True)
+        if min_count > 0 and not add_unk:
+            raise ValueError("Cannot require a minimum count if no unk-token is created.")
+
+        label_dictionary: Dictionary = Dictionary(add_unk=add_unk)
         label_dictionary.span_labels = False
 
         assert self.train
@@ -1448,7 +1429,6 @@ class Corpus:
         label_value_counter: typing.Counter[str] = Counter()
         all_sentence_labels: List[str] = []
         for sentence in Tqdm.tqdm(_iter_dataset(data)):
-
             # count all label types per sentence
             sentence_label_type_counter.update(sentence.annotation_layers.keys())
 
@@ -1496,6 +1476,61 @@ class Corpus:
 
         return label_dictionary
 
+    def _corrupt_labels(self, noise_share: float, label_type: str, labels: List[str], split: str = "train"):
+        """
+        Generates uniform label noise distribution in the chosen dataset split.
+        :noise_share: the desired share of noise in the train split.
+        :label_type: the type of labels for which the noise should be simulated.
+        :labels: an array with unique labels of said type (retrievable from label dictionary).
+        :split: in which dataset split the noise is to be simulated.
+        """
+        import numpy as np
+
+        if noise_share < 0 or noise_share > 1:
+            raise ValueError("noise_share must be between 0 and 1.")
+
+        if split == "train":
+            assert self.train
+            datasets = [self.train]
+        elif split == "dev":
+            assert self.dev
+            datasets = [self.dev]
+        elif split == "test":
+            assert self.test
+            datasets = [self.test]
+        else:
+            raise ValueError("split must be either train, dev or test.")
+
+        data: ConcatDataset = ConcatDataset(datasets)
+
+        orig_label_p = 1 - noise_share
+        other_label_p = noise_share / (len(labels) - 1)
+
+        corrupted_count = 0
+        total_label_count = 0
+
+        log.info("Generating noisy labels. Progress:")
+
+        for data_point in Tqdm.tqdm(_iter_dataset(data)):
+            for label in data_point.get_labels(label_type):
+                total_label_count += 1
+                orig_label = label.value
+                prob_dist = [other_label_p] * len(labels)
+                prob_dist[labels.index(orig_label)] = orig_label_p
+                # sample randomly from a label distribution according to the probabilities defined by the desired noise share
+                new_label = np.random.choice(a=labels, p=prob_dist)
+                # replace the old label with the new one
+                label.data_point.set_label(label_type, new_label)
+                # keep track of the old (clean) label using another label type category
+                label.data_point.add_label(label_type + "_clean", orig_label)
+                # keep track of how many labels in total are flipped
+                if new_label != orig_label:
+                    corrupted_count += 1
+
+        log.info(
+            f"Total labels corrupted: {corrupted_count}. Resulting noise share: {round((corrupted_count/total_label_count)*100, 2)}%."
+        )
+
     def get_label_distribution(self):
         class_to_count = defaultdict(lambda: 0)
         for sent in self.train:
@@ -1515,25 +1550,33 @@ class Corpus:
 
     @deprecated(version="0.8", reason="Use 'make_label_dictionary' instead.")
     def make_tag_dictionary(self, tag_type: str) -> Dictionary:
-
         # Make the tag dictionary
         tag_dictionary: Dictionary = Dictionary(add_unk=False)
         tag_dictionary.add_item("O")
         for sentence in _iter_dataset(self.get_all_sentences()):
             for token in sentence.tokens:
-                tag_dictionary.add_item(token.get_tag(tag_type).value)
+                tag_dictionary.add_item(token.get_label(tag_type).value)
         tag_dictionary.add_item("<START>")
         tag_dictionary.add_item("<STOP>")
         return tag_dictionary
 
 
 class MultiCorpus(Corpus):
-    def __init__(self, corpora: List[Corpus], name: str = "multicorpus", **corpusargs):
+    def __init__(
+        self,
+        corpora: List[Corpus],
+        task_ids: Optional[List[str]] = None,
+        name: str = "multicorpus",
+        **corpusargs,
+    ):
         self.corpora: List[Corpus] = corpora
 
+        ids = task_ids if task_ids else [f"Task_{i}" for i in range(len(corpora))]
+
         train_parts = []
         dev_parts = []
         test_parts = []
+        print(self.corpora)
         for corpus in self.corpora:
             if corpus.train:
                 train_parts.append(corpus.train)
@@ -1543,9 +1586,9 @@ class MultiCorpus(Corpus):
                 test_parts.append(corpus.test)
 
         super(MultiCorpus, self).__init__(
-            ConcatDataset(train_parts) if len(train_parts) > 0 else None,
-            ConcatDataset(dev_parts) if len(dev_parts) > 0 else None,
-            ConcatDataset(test_parts) if len(test_parts) > 0 else None,
+            ConcatFlairDataset(train_parts, ids) if len(train_parts) > 0 else None,
+            ConcatFlairDataset(dev_parts, ids) if len(dev_parts) > 0 else None,
+            ConcatFlairDataset(test_parts, ids) if len(test_parts) > 0 else None,
             name=name,
             **corpusargs,
         )
@@ -1561,6 +1604,63 @@ class MultiCorpus(Corpus):
         return output
 
 
+class FlairDataset(Dataset):
+    @abstractmethod
+    def is_in_memory(self) -> bool:
+        pass
+
+
+class ConcatFlairDataset(Dataset):
+    r"""Dataset as a concatenation of multiple datasets.
+
+    This class is useful to assemble different existing datasets.
+
+    Args:
+        datasets (sequence): List of datasets to be concatenated
+    """
+    datasets: List[Dataset]
+    cumulative_sizes: List[int]
+
+    @staticmethod
+    def cumsum(sequence):
+        r, s = [], 0
+        for e in sequence:
+            length_of_e = len(e)
+            r.append(length_of_e + s)
+            s += length_of_e
+        return r
+
+    def __init__(self, datasets: Iterable[Dataset], ids: Iterable[str]) -> None:
+        super(ConcatFlairDataset, self).__init__()
+        self.datasets = list(datasets)
+        self.ids = list(ids)
+        assert len(self.datasets) > 0, "datasets should not be an empty iterable"  # type: ignore[arg-type]
+        for d in self.datasets:
+            assert not isinstance(d, IterableDataset), "ConcatSentenceDataset does not support IterableDataset"
+        self.cumulative_sizes = self.cumsum(self.datasets)
+
+    def __len__(self):
+        return self.cumulative_sizes[-1]
+
+    def __getitem__(self, idx):
+        if idx < 0:
+            if -idx > len(self):
+                raise ValueError("absolute value of index should not exceed dataset length")
+            idx = len(self) + idx
+        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
+        if dataset_idx == 0:
+            sample_idx = idx
+        else:
+            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
+        sentence = self.datasets[dataset_idx][sample_idx]
+        sentence.set_label("multitask_id", self.ids[dataset_idx])
+        return sentence
+
+    @property
+    def cummulative_sizes(self):
+        return self.cumulative_sizes
+
+
 def iob2(tags):
     """
     Check that tags have a valid IOB format.
@@ -1583,30 +1683,6 @@ def iob2(tags):
     return True
 
 
-def iob_iobes(tags):
-    """
-    IOB -> IOBES
-    """
-    for i, tag in enumerate(tags):
-        if tag.value == "O" or tag.value == "":
-            tag.value = "O"
-            continue
-        t, label = tag.value.split("-", 1)
-        if len(tags) == i + 1 or tags[i + 1].value == "O":
-            next_same = False
-        else:
-            nt, next_label = tags[i + 1].value.split("-", 1)
-            next_same = nt == "I" and next_label == label
-        if t == "B":
-            if not next_same:
-                tag.value = tag.value.replace("B-", "S-")
-        elif t == "I":
-            if not next_same:
-                tag.value = tag.value.replace("I-", "E-")
-        else:
-            raise Exception("Invalid IOB format!")
-
-
 def randomly_split_into_two_datasets(dataset, length_of_first):
     import random
 
@@ -1619,3 +1695,62 @@ def randomly_split_into_two_datasets(dataset, length_of_first):
     second_dataset.sort()
 
     return Subset(dataset, first_dataset), Subset(dataset, second_dataset)
+
+
+def get_spans_from_bio(bioes_tags: List[str], bioes_scores=None) -> List[typing.Tuple[List[int], float, str]]:
+    # add a dummy "O" to close final prediction
+    bioes_tags.append("O")
+    # return complex list
+    found_spans = []
+    # internal variables
+    current_tag_weights: Dict[str, float] = defaultdict(lambda: 0.0)
+    previous_tag = "O-"
+    current_span: List[int] = []
+    current_span_scores: List[float] = []
+    for idx, bioes_tag in enumerate(bioes_tags):
+        # non-set tags are OUT tags
+        if bioes_tag == "" or bioes_tag == "O" or bioes_tag == "_":
+            bioes_tag = "O-"
+
+        # anything that is not OUT is IN
+        in_span = False if bioes_tag == "O-" else True
+
+        # does this prediction start a new span?
+        starts_new_span = False
+
+        # begin and single tags start new spans
+        if bioes_tag[0:2] in ["B-", "S-"]:
+            starts_new_span = True
+
+        # in IOB format, an I tag starts a span if it follows an O or is a different span
+        if bioes_tag[0:2] == "I-" and previous_tag[2:] != bioes_tag[2:]:
+            starts_new_span = True
+
+        # single tags that change prediction start new spans
+        if bioes_tag[0:2] in ["S-"] and previous_tag[2:] != bioes_tag[2:]:
+            starts_new_span = True
+
+        # if an existing span is ended (either by reaching O or starting a new span)
+        if (starts_new_span or not in_span) and len(current_span) > 0:
+            # determine score and value
+            span_score = sum(current_span_scores) / len(current_span_scores)
+            span_value = max(current_tag_weights.keys(), key=current_tag_weights.__getitem__)
+
+            # append to result list
+            found_spans.append((current_span, span_score, span_value))
+
+            # reset for-loop variables for new span
+            current_span = []
+            current_span_scores = []
+            current_tag_weights = defaultdict(lambda: 0.0)
+
+        if in_span:
+            current_span.append(idx)
+            current_span_scores.append(bioes_scores[idx] if bioes_scores else 1.0)
+            weight = 1.1 if starts_new_span else 1.0
+            current_tag_weights[bioes_tag[2:]] += weight
+
+        # remember previous tag
+        previous_tag = bioes_tag
+
+    return found_spans
