@@ -1,5 +1,5 @@
 import logging
-from typing import List, Union
+from typing import Any, Dict, List, Optional, Union, cast
 
 import torch
 from sklearn.feature_extraction.text import TfidfVectorizer
@@ -8,22 +8,25 @@ from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
 
 import flair
 from flair.data import Sentence
-from flair.embeddings.base import Embeddings, TransformerEmbedding
+from flair.embeddings.base import (
+    DocumentEmbeddings,
+    load_embeddings,
+    register_embeddings,
+)
 from flair.embeddings.token import FlairEmbeddings, StackedEmbeddings, TokenEmbeddings
+from flair.embeddings.transformer import (
+    TransformerEmbeddings,
+    TransformerOnnxDocumentEmbeddings,
+)
 from flair.nn import LockedDropout, WordDropout
 
 log = logging.getLogger("flair")
 
 
-class DocumentEmbeddings(Embeddings[Sentence]):
-    """Abstract base class for all document-level embeddings. Every new type of document embedding must implement these methods."""
+@register_embeddings
+class TransformerDocumentEmbeddings(DocumentEmbeddings, TransformerEmbeddings):
+    onnx_cls = TransformerOnnxDocumentEmbeddings
 
-    @property
-    def embedding_type(self) -> str:
-        return "sentence-level"
-
-
-class TransformerDocumentEmbeddings(DocumentEmbeddings, TransformerEmbedding):
     def __init__(
         self,
         model: str = "bert-base-uncased",  # set parameters with different default values
@@ -41,7 +44,7 @@ class TransformerDocumentEmbeddings(DocumentEmbeddings, TransformerEmbedding):
         :param layer_mean: If True, uses a scalar mix of layers as embedding
         :param fine_tune: If True, allows transformers to be fine-tuned during training
         """
-        TransformerEmbedding.__init__(
+        TransformerEmbeddings.__init__(
             self,
             model=model,
             layers=layers,
@@ -58,6 +61,7 @@ class TransformerDocumentEmbeddings(DocumentEmbeddings, TransformerEmbedding):
         return cls(**state)
 
 
+@register_embeddings
 class DocumentPoolEmbeddings(DocumentEmbeddings):
     def __init__(
         self,
@@ -73,7 +77,9 @@ class DocumentPoolEmbeddings(DocumentEmbeddings):
         """
         super().__init__()
 
-        if isinstance(embeddings, TokenEmbeddings):
+        if isinstance(embeddings, StackedEmbeddings):
+            embeddings = embeddings.embeddings
+        elif isinstance(embeddings, TokenEmbeddings):
             embeddings = [embeddings]
 
         self.embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embeddings)
@@ -99,6 +105,8 @@ class DocumentPoolEmbeddings(DocumentEmbeddings):
         self.pooling = pooling
         self.name: str = f"document_{self.pooling}"
 
+        self.eval()
+
     @property
     def embedding_length(self) -> int:
         return self.__embedding_length
@@ -114,7 +122,6 @@ class DocumentPoolEmbeddings(DocumentEmbeddings):
         self.embeddings.embed(sentences)
 
         for sentence in sentences:
-
             word_embeddings = torch.cat([token.get_embedding().unsqueeze(0) for token in sentence.tokens], dim=0).to(
                 flair.device
             )
@@ -141,11 +148,25 @@ class DocumentPoolEmbeddings(DocumentEmbeddings):
     def extra_repr(self):
         return f"fine_tune_mode={self.fine_tune_mode}, pooling={self.pooling}"
 
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "DocumentPoolEmbeddings":
+        embeddings = cast(StackedEmbeddings, load_embeddings(params.pop("embeddings"))).embeddings
+        return cls(embeddings=embeddings, **params)
 
+    def to_params(self) -> Dict[str, Any]:
+        return {
+            "pooling": self.pooling,
+            "fine_tune_mode": self.fine_tune_mode,
+            "embeddings": self.embeddings.save_embeddings(False),
+        }
+
+
+@register_embeddings
 class DocumentTFIDFEmbeddings(DocumentEmbeddings):
     def __init__(
         self,
-        train_dataset,
+        train_dataset: List[Sentence],
+        vectorizer: Optional[TfidfVectorizer] = None,
         **vectorizer_params,
     ):
         """The constructor for DocumentTFIDFEmbeddings.
@@ -156,14 +177,20 @@ class DocumentTFIDFEmbeddings(DocumentEmbeddings):
 
         import numpy as np
 
-        self.vectorizer = TfidfVectorizer(dtype=np.float32, **vectorizer_params)
-        self.vectorizer.fit([s.to_original_text() for s in train_dataset])
+        if vectorizer is not None:
+            self.vectorizer = vectorizer
+            if len(train_dataset) > 0:
+                raise ValueError("Cannot initialize document tfidf embeddings with a vectorizer and with a dataset")
+        else:
+            self.vectorizer = TfidfVectorizer(dtype=np.float32, **vectorizer_params)
+            self.vectorizer.fit([s.to_original_text() for s in train_dataset])
 
         self.__embedding_length: int = len(self.vectorizer.vocabulary_)
 
         self.to(flair.device)
 
         self.name: str = "document_tfidf"
+        self.eval()
 
     @property
     def embedding_length(self) -> int:
@@ -185,7 +212,17 @@ class DocumentTFIDFEmbeddings(DocumentEmbeddings):
     def _add_embeddings_internal(self, sentences: List[Sentence]):
         pass
 
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "DocumentTFIDFEmbeddings":
+        return cls(train_dataset=[], vectorizer=params["vectorizer"])
 
+    def to_params(self) -> Dict[str, Any]:
+        return {
+            "vectorizer": self.vectorizer,
+        }
+
+
+@register_embeddings
 class DocumentRNNEmbeddings(DocumentEmbeddings):
     def __init__(
         self,
@@ -284,7 +321,6 @@ class DocumentRNNEmbeddings(DocumentEmbeddings):
             self.word_dropout = None
 
         self.rnn.zero_grad()
-
         # embed words in the sentence
         self.embeddings.embed(sentences)
 
@@ -353,7 +389,6 @@ class DocumentRNNEmbeddings(DocumentEmbeddings):
             sentence.set_embedding(self.name, embedding)
 
     def _apply(self, fn):
-
         # models that were serialized using torch versions older than 1.4.0 lack the _flat_weights_names attribute
         # check if this is the case and if so, set it
         for child_module in self.children():
@@ -377,12 +412,10 @@ class DocumentRNNEmbeddings(DocumentEmbeddings):
 
             child_module._apply(fn)
 
-    def __getstate__(self):
-
+    def to_params(self):
         # serialize the language models and the constructor arguments (but nothing else)
         model_state = {
-            "state_dict": self.state_dict(),
-            "embeddings": self.embeddings.embeddings,
+            "embeddings": self.embeddings.save_embeddings(False),
             "hidden_size": self.rnn.hidden_size,
             "rnn_layers": self.rnn.num_layers,
             "reproject_words": self.reproject_words,
@@ -397,40 +430,54 @@ class DocumentRNNEmbeddings(DocumentEmbeddings):
 
         return model_state
 
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "DocumentRNNEmbeddings":
+        stacked_embeddings = load_embeddings(params["embeddings"])
+        assert isinstance(stacked_embeddings, StackedEmbeddings)
+        return cls(
+            embeddings=stacked_embeddings.embeddings,
+            hidden_size=params["hidden_size"],
+            rnn_layers=params["rnn_layers"],
+            reproject_words=params["reproject_words"],
+            reproject_words_dimension=params["reproject_words_dimension"],
+            bidirectional=params["bidirectional"],
+            dropout=params["dropout"],
+            word_dropout=params["word_dropout"],
+            locked_dropout=params["locked_dropout"],
+            rnn_type=params["rnn_type"],
+            fine_tune=params["fine_tune"],
+        )
+
     def __setstate__(self, d):
+        # re-initialize language model with constructor arguments
+        language_model = DocumentRNNEmbeddings(
+            embeddings=d["embeddings"],
+            hidden_size=d["hidden_size"],
+            rnn_layers=d["rnn_layers"],
+            reproject_words=d["reproject_words"],
+            reproject_words_dimension=d["reproject_words_dimension"],
+            bidirectional=d["bidirectional"],
+            dropout=d["dropout"],
+            word_dropout=d["word_dropout"],
+            locked_dropout=d["locked_dropout"],
+            rnn_type=d["rnn_type"],
+            fine_tune=d["fine_tune"],
+        )
 
         # special handling for deserializing language models
         if "state_dict" in d:
-
-            # re-initialize language model with constructor arguments
-            language_model = DocumentRNNEmbeddings(
-                embeddings=d["embeddings"],
-                hidden_size=d["hidden_size"],
-                rnn_layers=d["rnn_layers"],
-                reproject_words=d["reproject_words"],
-                reproject_words_dimension=d["reproject_words_dimension"],
-                bidirectional=d["bidirectional"],
-                dropout=d["dropout"],
-                word_dropout=d["word_dropout"],
-                locked_dropout=d["locked_dropout"],
-                rnn_type=d["rnn_type"],
-                fine_tune=d["fine_tune"],
-            )
-
             language_model.load_state_dict(d["state_dict"])
 
-            # copy over state dictionary to self
-            for key in language_model.__dict__.keys():
-                self.__dict__[key] = language_model.__dict__[key]
+        # copy over state dictionary to self
+        for key in language_model.__dict__.keys():
+            self.__dict__[key] = language_model.__dict__[key]
 
-            # set the language model to eval() by default (this is necessary since FlairEmbeddings "protect" the LM
-            # in their "self.train()" method)
-            self.eval()
-
-        else:
-            self.__dict__ = d
+        # set the language model to eval() by default (this is necessary since FlairEmbeddings "protect" the LM
+        # in their "self.train()" method)
+        self.eval()
 
 
+@register_embeddings
 class DocumentLMEmbeddings(DocumentEmbeddings):
     def __init__(self, flair_embeddings: List[FlairEmbeddings]):
         super().__init__()
@@ -445,19 +492,18 @@ class DocumentLMEmbeddings(DocumentEmbeddings):
                 self.static_embeddings = False
 
         self._embedding_length: int = sum(embedding.embedding_length for embedding in flair_embeddings)
+        self.eval()
 
     @property
     def embedding_length(self) -> int:
         return self._embedding_length
 
     def _add_embeddings_internal(self, sentences: List[Sentence]):
-
         for embedding in self.embeddings:
             embedding.embed(sentences)
 
             # iterate over sentences
             for sentence in sentences:
-
                 # if its a forward LM, take last state
                 if embedding.is_forward_lm:
                     sentence.set_embedding(
@@ -469,19 +515,31 @@ class DocumentLMEmbeddings(DocumentEmbeddings):
 
         return sentences
 
+    def get_names(self) -> List[str]:
+        if "__names" not in self.__dict__:
+            self.__names = [name for embedding in self.embeddings for name in embedding.get_names()]
 
+        return self.__names
+
+    def to_params(self) -> Dict[str, Any]:
+        return {"flair_embeddings": [embedding.save_embeddings(False) for embedding in self.embeddings]}
+
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "DocumentLMEmbeddings":
+        return cls([cast(FlairEmbeddings, load_embeddings(embedding)) for embedding in params["flair_embeddings"]])
+
+
+@register_embeddings
 class SentenceTransformerDocumentEmbeddings(DocumentEmbeddings):
     def __init__(
         self,
         model: str = "bert-base-nli-mean-tokens",
         batch_size: int = 1,
-        convert_to_numpy: bool = False,
     ):
         """
         :param model: string name of models from SentencesTransformer Class
         :param name: string name of embedding type which will be set to Sentence object
         :param batch_size: int number of sentences to processed in one batch
-        :param convert_to_numpy: bool whether the encode() returns a numpy array or PyTorch tensor
         """
         super().__init__()
 
@@ -494,14 +552,16 @@ class SentenceTransformerDocumentEmbeddings(DocumentEmbeddings):
             log.warning("-" * 100)
             pass
 
-        self.model = SentenceTransformer(model)
+        self.model_name = model
+        self.model = SentenceTransformer(
+            model, cache_folder=str(flair.cache_root / "embeddings" / "sentence-transformer")
+        )
         self.name = "sentence-transformers-" + str(model)
         self.batch_size = batch_size
-        self.convert_to_numpy = convert_to_numpy
         self.static_embeddings = True
+        self.eval()
 
     def _add_embeddings_internal(self, sentences: List[Sentence]) -> List[Sentence]:
-
         sentence_batches = [
             sentences[i * self.batch_size : (i + 1) * self.batch_size]
             for i in range((len(sentences) + self.batch_size - 1) // self.batch_size)
@@ -513,11 +573,10 @@ class SentenceTransformerDocumentEmbeddings(DocumentEmbeddings):
         return sentences
 
     def _add_embeddings_to_sentences(self, sentences: List[Sentence]):
-
         # convert to plain strings, embedded in a list for the encode function
         sentences_plain_text = [sentence.to_plain_string() for sentence in sentences]
 
-        embeddings = self.model.encode(sentences_plain_text, convert_to_numpy=self.convert_to_numpy)
+        embeddings = self.model.encode(sentences_plain_text, convert_to_numpy=False)
         for sentence, embedding in zip(sentences, embeddings):
             sentence.set_embedding(self.name, embedding)
 
@@ -526,7 +585,18 @@ class SentenceTransformerDocumentEmbeddings(DocumentEmbeddings):
         """Returns the length of the embedding vector."""
         return self.model.get_sentence_embedding_dimension()
 
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "SentenceTransformerDocumentEmbeddings":
+        return cls(**params)
 
+    def to_params(self) -> Dict[str, Any]:
+        return {
+            "model": self.model_name,
+            "batch_size": self.batch_size,
+        }
+
+
+@register_embeddings
 class DocumentCNNEmbeddings(DocumentEmbeddings):
     def __init__(
         self,
@@ -564,7 +634,13 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
         if self.reproject_words and reproject_words_dimension is not None:
             self.embeddings_dimension = reproject_words_dimension
 
-        self.word_reprojection_map = torch.nn.Linear(self.length_of_all_token_embeddings, self.embeddings_dimension)
+        if self.reproject_words:
+            self.word_reprojection_map: Optional[torch.nn.Linear] = torch.nn.Linear(
+                self.length_of_all_token_embeddings, self.embeddings_dimension
+            )
+            torch.nn.init.xavier_uniform_(self.word_reprojection_map.weight)
+        else:
+            self.word_reprojection_map = None
 
         # CNN
         self.__embedding_length: int = sum([kernel_num for kernel_num, kernel_size in self.kernels])
@@ -583,10 +659,8 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
         self.locked_dropout = LockedDropout(locked_dropout) if locked_dropout > 0.0 else None
         self.word_dropout = WordDropout(word_dropout) if word_dropout > 0.0 else None
 
-        torch.nn.init.xavier_uniform_(self.word_reprojection_map.weight)
-
         self.to(flair.device)
-
+        self.min_sequence_length = max(kernel_size for _, kernel_size in self.kernels)
         self.eval()
 
     @property
@@ -609,10 +683,10 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
         self.embeddings.embed(sentences)
 
         lengths: List[int] = [len(sentence.tokens) for sentence in sentences]
-        longest_token_sequence_in_batch: int = max(lengths)
+        padding_length: int = max(max(lengths), self.min_sequence_length)
 
         pre_allocated_zero_tensor = torch.zeros(
-            self.embeddings.embedding_length * longest_token_sequence_in_batch,
+            self.embeddings.embedding_length * padding_length,
             dtype=torch.float,
             device=flair.device,
         )
@@ -620,7 +694,7 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
         all_embs: List[torch.Tensor] = list()
         for sentence in sentences:
             all_embs += [emb for token in sentence for emb in token.get_each_embedding()]
-            nb_padding_tokens = longest_token_sequence_in_batch - len(sentence)
+            nb_padding_tokens = padding_length - len(sentence)
 
             if nb_padding_tokens > 0:
                 t = pre_allocated_zero_tensor[: self.embeddings.embedding_length * nb_padding_tokens]
@@ -629,7 +703,7 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
         sentence_tensor = torch.cat(all_embs).view(
             [
                 len(sentences),
-                longest_token_sequence_in_batch,
+                padding_length,
                 self.embeddings.embedding_length,
             ]
         )
@@ -643,7 +717,7 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
             sentence_tensor = self.word_dropout(sentence_tensor)
 
         # reproject if set
-        if self.reproject_words:
+        if self.word_reprojection_map is not None:
             sentence_tensor = self.word_reprojection_map(sentence_tensor)
 
         # push CNN
@@ -674,3 +748,20 @@ class DocumentCNNEmbeddings(DocumentEmbeddings):
     def _apply(self, fn):
         for child_module in self.children():
             child_module._apply(fn)
+
+    @classmethod
+    def from_params(cls, params: Dict[str, Any]) -> "DocumentCNNEmbeddings":
+        embeddings = cast(StackedEmbeddings, load_embeddings(params.pop("embeddings"))).embeddings
+        return cls(embeddings=embeddings, **params)
+
+    def to_params(self) -> Dict[str, Any]:
+        return {
+            "embeddings": self.embeddings.save_embeddings(False),
+            "kernels": self.kernels,
+            "reproject_words": self.reproject_words,
+            "reproject_words_dimension": self.embeddings_dimension,
+            "dropout": 0.0 if self.dropout is None else self.dropout.p,
+            "word_dropout": 0.0 if self.word_dropout is None else self.word_dropout.p,
+            "locked_dropout": 0.0 if self.locked_dropout is None else self.locked_dropout.dropout_rate,
+            "fine_tune": not self.static_embeddings,
+        }
