@@ -1,30 +1,33 @@
 from .clustering import ClusteringModel
-from .dependency_parser_model import DependencyParser
 from .entity_linker_model import EntityLinker
 from .language_model import LanguageModel
 from .lemmatizer_model import Lemmatizer
+from .multitask_model import MultitaskModel
 from .pairwise_classification_model import TextPairClassifier
 from .regexp_tagger import RegexpTagger
+from .relation_classifier_model import RelationClassifier
 from .relation_extractor_model import RelationExtractor
-from .sequence_tagger_model import MultiTagger, SequenceTagger
+from .sequence_tagger_model import SequenceTagger
 from .tars_model import FewshotClassifier, TARSClassifier, TARSTagger
 from .text_classification_model import TextClassifier
+from .text_regression_model import TextRegressor
 from .word_tagger_model import WordTagger
 
 __all__ = [
-    "DependencyParser",
     "EntityLinker",
     "LanguageModel",
     "Lemmatizer",
     "TextPairClassifier",
+    "RelationClassifier",
     "RelationExtractor",
     "RegexpTagger",
-    "MultiTagger",
     "SequenceTagger",
     "WordTagger",
     "FewshotClassifier",
     "TARSClassifier",
     "TARSTagger",
     "TextClassifier",
+    "TextRegressor",
     "ClusteringModel",
+    "MultitaskModel",
 ]
