@@ -1,6 +1,7 @@
 import logging
+import typing
 from pathlib import Path
-from typing import List, Optional, Tuple, Union
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import torch
 import torch.nn as nn
@@ -9,22 +10,22 @@ from tqdm import tqdm
 
 import flair
 import flair.embeddings
-from flair.data import Sentence
+from flair.data import Corpus, Dictionary, Sentence, _iter_dataset
 from flair.datasets import DataLoader, FlairDatapointDataset
+from flair.embeddings.base import load_embeddings
+from flair.nn.model import ReduceTransformerVocabMixin
 from flair.training_utils import MetricRegression, Result, store_embeddings
 
 log = logging.getLogger("flair")
 
 
-class TextRegressor(flair.nn.Model[Sentence]):
+class TextRegressor(flair.nn.Model[Sentence], ReduceTransformerVocabMixin):
     def __init__(
         self,
         document_embeddings: flair.embeddings.DocumentEmbeddings,
         label_name: str = "label",
     ):
-
         super().__init__()
-        log.info("Using REGRESSION - experimental")
 
         self.document_embeddings: flair.embeddings.DocumentEmbeddings = document_embeddings
         self.label_name = label_name
@@ -33,37 +34,35 @@ class TextRegressor(flair.nn.Model[Sentence]):
 
         nn.init.xavier_uniform_(self.decoder.weight)
 
-        self.loss_function = nn.MSELoss()
+        self.loss_function = nn.MSELoss(reduction="sum")
 
         # auto-spawn on GPU if available
         self.to(flair.device)
 
+    @property
     def label_type(self):
         return self.label_name
 
-    def forward(self, sentences):
-
+    def _prepare_tensors(self, sentences: List[Sentence]) -> Tuple[torch.Tensor]:
         self.document_embeddings.embed(sentences)
-
         embedding_names = self.document_embeddings.get_names()
-
         text_embedding_list = [sentence.get_embedding(embedding_names).unsqueeze(0) for sentence in sentences]
         text_embedding_tensor = torch.cat(text_embedding_list, 0).to(flair.device)
+        return (text_embedding_tensor,)
 
+    def forward(self, *args: torch.Tensor) -> torch.Tensor:
+        (text_embedding_tensor,) = args
         label_scores = self.decoder(text_embedding_tensor)
-
         return label_scores
 
-    def forward_loss(self, data_points: Union[List[Sentence], Sentence]) -> torch.Tensor:
+    def forward_loss(self, sentences: List[Sentence]) -> Tuple[torch.Tensor, int]:
+        labels = self._labels_to_tensor(sentences)
+        text_embedding_tensor = self._prepare_tensors(sentences)
+        scores = self.forward(*text_embedding_tensor)
 
-        if not isinstance(data_points, list):
-            data_points = [data_points]
+        return self.loss_function(scores.squeeze(1), labels), len(sentences)
 
-        scores = self.forward(data_points)
-
-        return self._calculate_loss(scores, data_points)
-
-    def _labels_to_indices(self, sentences: List[Sentence]):
+    def _labels_to_tensor(self, sentences: List[Sentence]):
         indices = [
             torch.tensor([float(label.value) for label in sentence.labels], dtype=torch.float) for sentence in sentences
         ]
@@ -80,7 +79,6 @@ class TextRegressor(flair.nn.Model[Sentence]):
         label_name: Optional[str] = None,
         embedding_storage_mode="none",
     ) -> List[Sentence]:
-
         if label_name is None:
             label_name = self.label_name if self.label_name is not None else "label"
 
@@ -91,7 +89,9 @@ class TextRegressor(flair.nn.Model[Sentence]):
             if not sentences:
                 return sentences
 
-            reordered_sentences = sorted(sentences, key=lambda s: len(s), reverse=True)
+            Sentence.set_context_for_sentences(sentences)
+            filtered_sentences = self._filter_empty_sentences(sentences)
+            reordered_sentences = sorted(filtered_sentences, key=lambda s: len(s), reverse=True)
 
             if len(reordered_sentences) == 0:
                 return sentences
@@ -110,9 +110,11 @@ class TextRegressor(flair.nn.Model[Sentence]):
                 # stop if all sentences are empty
                 if not batch:
                     continue
-                scores = self.forward(batch)
 
-                for (sentence, score) in zip(batch, scores.tolist()):
+                (sentence_tensor,) = self._prepare_tensors(batch)
+                scores = self.forward(sentence_tensor)
+
+                for sentence, score in zip(batch, scores.tolist()):
                     sentence.set_label(label_name, value=str(score[0]))
 
                 # clearing token embeddings to save memory
@@ -120,24 +122,12 @@ class TextRegressor(flair.nn.Model[Sentence]):
 
             return sentences
 
-    def _calculate_loss(self, scores: torch.Tensor, sentences: List[Sentence]) -> torch.Tensor:
-        """
-        Calculates the loss.
-        :param scores: the prediction scores from the model
-        :param sentences: list of sentences
-        :return: loss value
-        """
-        return self.loss_function(scores.squeeze(1), self._labels_to_indices(sentences))
+    def forward_labels_and_loss(self, sentences: List[Sentence]) -> Tuple[torch.Tensor, torch.Tensor]:
+        labels = self._labels_to_tensor(sentences)
+        text_embedding_tensor = self._prepare_tensors(sentences)
+        scores = self.forward(*text_embedding_tensor)
 
-    def forward_labels_and_loss(
-        self, sentences: Union[Sentence, List[Sentence]]
-    ) -> Tuple[List[List[float]], torch.Tensor]:
-        if not isinstance(sentences, list):
-            sentences = [sentences]
-
-        scores = self.forward(sentences)
-        loss = self._calculate_loss(scores, sentences)
-        return scores, loss
+        return scores, self.loss_function(scores.squeeze(1), labels)
 
     def evaluate(
         self,
@@ -147,9 +137,12 @@ class TextRegressor(flair.nn.Model[Sentence]):
         embedding_storage_mode: str = "none",
         mini_batch_size: int = 32,
         num_workers: Optional[int] = 8,
+        main_evaluation_metric: Tuple[str, str] = ("micro avg", "f1-score"),
+        exclude_labels: List[str] = [],
+        gold_label_dictionary: Optional[Dictionary] = None,
+        return_loss: bool = True,
         **kwargs,
     ) -> Result:
-
         # read Dataset into data loader, if list of sentences passed, make Dataset first
         if not isinstance(data_points, Dataset):
             data_points = FlairDatapointDataset(data_points)
@@ -162,8 +155,7 @@ class TextRegressor(flair.nn.Model[Sentence]):
 
             lines: List[str] = []
             total_count = 0
-            for batch_nr, batch in enumerate(data_loader):
-
+            for batch in data_loader:
                 if isinstance(batch, Sentence):
                     batch = [batch]
 
@@ -175,9 +167,7 @@ class TextRegressor(flair.nn.Model[Sentence]):
                     for label in sentence.get_labels(gold_label_type):
                         true_values.append(float(label.value))
 
-                results = []
-                for score in scores:
-                    results.append(score[0])
+                results = scores[:, 0].cpu().tolist()
 
                 eval_loss += loss
 
@@ -196,7 +186,6 @@ class TextRegressor(flair.nn.Model[Sentence]):
             if out_path is not None:
                 with open(out_path, "w", encoding="utf-8") as outfile:
                     outfile.write("".join(lines))
-
             log_line = f"{metric.mean_squared_error()}\t{metric.spearmanr()}" f"\t{metric.pearsonr()}"
             log_header = "MSE\tSPEARMAN\tPEARSON"
 
@@ -220,17 +209,18 @@ class TextRegressor(flair.nn.Model[Sentence]):
     def _get_state_dict(self):
         model_state = {
             **super()._get_state_dict(),
-            "document_embeddings": self.document_embeddings,
+            "document_embeddings": self.document_embeddings.save_embeddings(use_state_dict=False),
             "label_name": self.label_type,
         }
         return model_state
 
     @classmethod
     def _init_model_with_state_dict(cls, state, **kwargs):
-        label_name = state["label_name"] if "label_name" in state.keys() else None
-
+        embeddings = state["document_embeddings"]
+        if isinstance(embeddings, dict):
+            embeddings = load_embeddings(embeddings)
         return super()._init_model_with_state_dict(
-            state, document_embeddings=state["document_embeddings"], label_name=label_name, **kwargs
+            state, document_embeddings=embeddings, label_name=state.get("label_name"), **kwargs
         )
 
     @staticmethod
@@ -239,3 +229,13 @@ class TextRegressor(flair.nn.Model[Sentence]):
         if len(sentences) != len(filtered_sentences):
             log.warning("Ignore {} sentence(s) with no tokens.".format(len(sentences) - len(filtered_sentences)))
         return filtered_sentences
+
+    @classmethod
+    def load(cls, model_path: Union[str, Path, Dict[str, Any]]) -> "TextRegressor":
+        from typing import cast
+
+        return cast("TextRegressor", super().load(model_path=model_path))
+
+    def get_used_tokens(self, corpus: Corpus) -> typing.Iterable[List[str]]:
+        for sentence in _iter_dataset(corpus.get_all_sentences()):
+            yield [t.text for t in sentence]
