@@ -1,8 +1,9 @@
+import inspect
 import itertools
 import logging
 import typing
 import warnings
-from abc import abstractmethod
+from abc import ABC, abstractmethod
 from collections import Counter
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Set, Tuple, Union
@@ -13,16 +14,17 @@ from torch.utils.data.dataset import Dataset
 from tqdm import tqdm
 
 import flair
-from flair import file_utils
-from flair.data import DT, DataPoint, Dictionary, Sentence
+from flair.data import DT, DT2, Corpus, Dictionary, Sentence, _iter_dataset
 from flair.datasets import DataLoader, FlairDatapointDataset
-from flair.file_utils import Tqdm
+from flair.embeddings import Embeddings
+from flair.embeddings.base import load_embeddings
+from flair.file_utils import Tqdm, load_torch_state
 from flair.training_utils import Result, store_embeddings
 
 log = logging.getLogger("flair")
 
 
-class Model(torch.nn.Module, typing.Generic[DT]):
+class Model(torch.nn.Module, typing.Generic[DT], ABC):
     """Abstract base class for all downstream task models in Flair,
     such as SequenceTagger and TextClassifier.
     Every new type of model must implement these methods."""
@@ -37,7 +39,7 @@ class Model(torch.nn.Module, typing.Generic[DT]):
         raise NotImplementedError
 
     @abstractmethod
-    def forward_loss(self, data_points: Union[List[DT], DT]) -> Union[torch.Tensor, Tuple[torch.Tensor, int]]:
+    def forward_loss(self, data_points: List[DT]) -> Tuple[torch.Tensor, int]:
         """Performs a forward pass and returns a loss tensor for backpropagation.
         Implement this to enable training."""
         raise NotImplementedError
@@ -51,6 +53,10 @@ class Model(torch.nn.Module, typing.Generic[DT]):
         embedding_storage_mode: str = "none",
         mini_batch_size: int = 32,
         num_workers: Optional[int] = 8,
+        main_evaluation_metric: Tuple[str, str] = ("micro avg", "f1-score"),
+        exclude_labels: List[str] = [],
+        gold_label_dictionary: Optional[Dictionary] = None,
+        return_loss: bool = True,
         **kwargs,
     ) -> Result:
         """Evaluates the model. Returns a Result object containing evaluation
@@ -66,14 +72,24 @@ class Model(torch.nn.Module, typing.Generic[DT]):
         """Returns the state dictionary for this model."""
         state_dict = {"state_dict": self.state_dict()}
 
+        # Always include the name of the Model class for which the state dict holds
+        state_dict["__cls__"] = self.__class__.__name__
+
         return state_dict
 
     @classmethod
     def _init_model_with_state_dict(cls, state, **kwargs):
         """Initialize the model from a state dictionary."""
+        if "embeddings" in kwargs:
+            embeddings = kwargs.pop("embeddings")
+            if isinstance(embeddings, dict):
+                embeddings = load_embeddings(embeddings)
+            kwargs["embeddings"] = embeddings
+
         model = cls(**kwargs)
 
         model.load_state_dict(state["state_dict"])
+
         return model
 
     @staticmethod
@@ -92,7 +108,6 @@ class Model(torch.nn.Module, typing.Generic[DT]):
 
         # write out a "model card" if one is set
         if self.model_card is not None:
-
             # special handling for optimizer:
             # remember optimizer class and state dictionary
             if "training_parameters" in self.model_card:
@@ -125,29 +140,68 @@ class Model(torch.nn.Module, typing.Generic[DT]):
                 self.model_card["training_parameters"]["scheduler"] = scheduler
 
     @classmethod
-    def load(cls, model_path: Union[str, Path]):
+    def load(cls, model_path: Union[str, Path, Dict[str, Any]]) -> "Model":
         """
         Loads the model from the given file.
-        :param model_path: the model file
+        :param model_path: the model file or the already loaded state dict
         :return: the loaded text classifier model
         """
-        model_file = cls._fetch_model(str(model_path))
+        # if this class is abstract, go through all inheriting classes and try to fetch and load the model
+        if inspect.isabstract(cls):
+            # get all non-abstract subclasses
+            subclasses = get_non_abstract_subclasses(cls)
 
-        with warnings.catch_warnings():
-            warnings.filterwarnings("ignore")
-            # load_big_file is a workaround byhttps://github.com/highway11git
-            # to load models on some Mac/Windows setups
-            # see https://github.com/zalandoresearch/flair/issues/351
-            f = file_utils.load_big_file(str(model_file))
-            state = torch.load(f, map_location="cpu")
+            # try to fetch the model for each subclass. if fetching is possible, load model and return it
+            for model_cls in subclasses:
+                try:
+                    new_model_path = model_cls._fetch_model(model_path)
+                    if new_model_path != model_path:
+                        return model_cls.load(new_model_path)
+                except Exception:
+                    # skip any invalid loadings, e.g. not found on huggingface hub
+                    continue
 
-        model = cls._init_model_with_state_dict(state)
+            # if the model cannot be fetched, load as a file
+            state = model_path if isinstance(model_path, dict) else load_torch_state(str(model_path))
 
-        if "model_card" in state:
-            model.model_card = state["model_card"]
+            # try to get model class from state
+            cls_name = state.pop("__cls__", None)
+            if cls_name:
+                for model_cls in subclasses:
+                    if cls_name == model_cls.__name__:
+                        return model_cls.load(state)
 
-        model.eval()
-        model.to(flair.device)
+            # older (flair 11.3 and below) models do not contain cls information. In this case, try all subclasses
+            for model_cls in subclasses:
+                # if str(model_cls) == "<class 'flair.models.pairwise_classification_model.TextPairClassifier'>": continue
+                try:
+                    model = model_cls.load(state)
+                    return model
+                except Exception as e:
+                    print(e)
+                    # skip any invalid loadings, e.g. not found on huggingface hub
+                    continue
+
+            raise ValueError(f"Could not find any model with name '{model_path}'")
+
+        else:
+            # if this class is not abstract, fetch the model and load it
+            if not isinstance(model_path, dict):
+                model_file = cls._fetch_model(str(model_path))
+                state = load_torch_state(model_file)
+            else:
+                state = model_path
+
+            if "__cls__" in state:
+                state.pop("__cls__")
+
+            model = cls._init_model_with_state_dict(state)
+
+            if "model_card" in state:
+                model.model_card = state["model_card"]
+
+            model.eval()
+            model.to(flair.device)
 
         return model
 
@@ -180,7 +234,13 @@ class Model(torch.nn.Module, typing.Generic[DT]):
             )
 
 
-class Classifier(Model[DT], typing.Generic[DT]):
+class ReduceTransformerVocabMixin(ABC):
+    @abstractmethod
+    def get_used_tokens(self, corpus: Corpus) -> typing.Iterable[List[str]]:
+        pass
+
+
+class Classifier(Model[DT], typing.Generic[DT], ReduceTransformerVocabMixin, ABC):
     """Abstract base class for all Flair models that do classification,
     both single- and multi-label. It inherits from flair.nn.Model and adds an
     unified evaluate() function so that all classification models use the same
@@ -197,17 +257,21 @@ class Classifier(Model[DT], typing.Generic[DT]):
         main_evaluation_metric: Tuple[str, str] = ("micro avg", "f1-score"),
         exclude_labels: List[str] = [],
         gold_label_dictionary: Optional[Dictionary] = None,
+        return_loss: bool = True,
         **kwargs,
     ) -> Result:
         import numpy as np
         import sklearn
 
+        # make sure <unk> is contained in gold_label_dictionary, if given
+        if gold_label_dictionary and not gold_label_dictionary.add_unk:
+            raise AssertionError("gold_label_dictionary must have add_unk set to true in initialization.")
+
         # read Dataset into data loader, if list of sentences passed, make Dataset first
         if not isinstance(data_points, Dataset):
             data_points = FlairDatapointDataset(data_points)
 
         with torch.no_grad():
-
             # loss calculation
             eval_loss = torch.zeros(1, device=flair.device)
             average_over = 0
@@ -224,28 +288,28 @@ class Classifier(Model[DT], typing.Generic[DT]):
 
             sentence_id = 0
             for batch in Tqdm.tqdm(loader):
-
                 # remove any previously predicted labels
                 for datapoint in batch:
                     datapoint.remove_labels("predicted")
+
                 # predict for batch
                 loss_and_count = self.predict(
                     batch,
                     embedding_storage_mode=embedding_storage_mode,
                     mini_batch_size=mini_batch_size,
                     label_name="predicted",
-                    return_loss=True,
+                    return_loss=return_loss,
                 )
 
-                if isinstance(loss_and_count, tuple):
-                    average_over += loss_and_count[1]
-                    eval_loss += loss_and_count[0]
-                else:
-                    eval_loss += loss_and_count
+                if return_loss:
+                    if isinstance(loss_and_count, tuple):
+                        average_over += loss_and_count[1]
+                        eval_loss += loss_and_count[0]
+                    else:
+                        eval_loss += loss_and_count
 
                 # get the gold labels
                 for datapoint in batch:
-
                     for gold_label in datapoint.get_labels(gold_label_type):
                         representation = str(sentence_id) + ": " + gold_label.unlabeled_identifier
 
@@ -477,26 +541,36 @@ class Classifier(Model[DT], typing.Generic[DT]):
             correct_string = " -> MISMATCH!\n" if g != p else ""
             # print info
             eval_line = (
-                f"{datapoint.to_original_text()}\n"
+                f"{datapoint.text}\n"
                 f" - Gold: {', '.join(label.value if label.data_point == datapoint else label.labeled_identifier for label in datapoint.get_labels(gold_label_type))}\n"
                 f" - Pred: {', '.join(label.value if label.data_point == datapoint else label.labeled_identifier for label in datapoint.get_labels('predicted'))}\n{correct_string}\n"
             )
             lines.append(eval_line)
         return lines
 
+    def get_used_tokens(self, corpus: Corpus) -> typing.Iterable[List[str]]:
+        for sentence in _iter_dataset(corpus.get_all_sentences()):
+            yield [t.text for t in sentence]
 
-class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
+    @classmethod
+    def load(cls, model_path: Union[str, Path, Dict[str, Any]]) -> "Classifier":
+        from typing import cast
+
+        return cast("Classifier", super().load(model_path=model_path))
+
+
+class DefaultClassifier(Classifier[DT], typing.Generic[DT, DT2], ABC):
     """Default base class for all Flair models that do classification, both
     single- and multi-label. It inherits from flair.nn.Classifier and thus from
     flair.nn.Model. All features shared by all classifiers are implemented here,
     including the loss calculation and the predict() method. Currently, the
     TextClassifier, RelationExtractor, TextPairClassifier and
-    SimpleSequenceTagger implement this class. You only need to implement the
-    forward_pass() method to implement this base class.
+    SimpleSequenceTagger implement this class.
     """
 
     def __init__(
         self,
+        embeddings: Embeddings,
         label_dictionary: Dictionary,
         final_embedding_size: int,
         dropout: float = 0.0,
@@ -506,10 +580,15 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         multi_label_threshold: float = 0.5,
         loss_weights: Dict[str, float] = None,
         decoder: Optional[torch.nn.Module] = None,
+        inverse_model: bool = False,
+        train_on_gold_pairs_only: bool = False,
+        should_embed_sentence: bool = True,
     ):
-
         super().__init__()
 
+        # set the embeddings
+        self.embeddings = embeddings
+
         # initialize the label dictionary
         self.label_dictionary: Dictionary = label_dictionary
 
@@ -525,11 +604,14 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         # set up multi-label logic
         self.multi_label = multi_label
         self.multi_label_threshold = multi_label_threshold
+        self.final_embedding_size = final_embedding_size
+        self.inverse_model = inverse_model
 
         # init dropouts
         self.dropout: torch.nn.Dropout = torch.nn.Dropout(dropout)
         self.locked_dropout = flair.nn.LockedDropout(locked_dropout)
         self.word_dropout = flair.nn.WordDropout(word_dropout)
+        self.should_embed_sentence = should_embed_sentence
 
         # loss weights and loss function
         self.weight_dict = loss_weights
@@ -544,26 +626,46 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         else:
             self.loss_weights = None
 
-        if self.multi_label:
-            self.loss_function: _Loss = torch.nn.BCEWithLogitsLoss(weight=self.loss_weights)
-        else:
-            self.loss_function = torch.nn.CrossEntropyLoss(weight=self.loss_weights)
+        # set up gradient reversal if so specified
+        if inverse_model:
+            from pytorch_revgrad import RevGrad
 
-    def forward_pass(
-        self,
-        sentences: Union[List[DT], DT],
-        for_prediction: bool = False,
-    ) -> Union[Tuple[torch.Tensor, List[List[str]]], Tuple[torch.Tensor, List[List[str]], List[DataPoint]]]:
-        """This method does a forward pass through the model given a list of data
-        points as input.
-        Returns the tuple (embeddings, labels) if return_label_candidates = False,
-        where scores are a tensor of logits produced by the decoder and labels
-        are the string labels for each data point.
-        Returns the tuple (embeddings, labels, data_points) if return_label_candidates = True,
-        where data_points are the data points to which labels are added (commonly
-        either Sentence, Span or Token objects)."""
+            self.gradient_reversal = RevGrad()
+
+        if self.multi_label:
+            self.loss_function: _Loss = torch.nn.BCEWithLogitsLoss(weight=self.loss_weights, reduction="sum")
+        else:
+            self.loss_function = torch.nn.CrossEntropyLoss(weight=self.loss_weights, reduction="sum")
+        self.train_on_gold_pairs_only = train_on_gold_pairs_only
+
+    def _filter_data_point(self, data_point: DT) -> bool:
+        """Specify if a data point should be kept. That way you can remove for example empty texts.
+        Return true if the data point should be kept and false if it should be removed.
+        """
+        return True if len(data_point) > 0 else False
+
+    @abstractmethod
+    def _get_embedding_for_data_point(self, prediction_data_point: DT2) -> torch.Tensor:
+        raise NotImplementedError()
+
+    @abstractmethod
+    def _get_data_points_from_sentence(self, sentence: DT) -> List[DT2]:
+        """Returns the data_points to which labels are added (Sentence, Span, Token, ... objects)"""
         raise NotImplementedError
 
+    def _get_data_points_for_batch(self, sentences: List[DT]) -> List[DT2]:
+        """Returns the data_points to which labels are added (Sentence, Span, Token, ... objects)"""
+        return [data_point for sentence in sentences for data_point in self._get_data_points_from_sentence(sentence)]
+
+    def _get_label_of_datapoint(self, data_point: DT2) -> List[str]:
+        """Extracts the labels from the data points.
+        Each data point might return a list of strings, representing multiple labels.
+        """
+        if self.multi_label:
+            return [label.value for label in data_point.get_labels(self.label_type)]
+        else:
+            return [data_point.get_label(self.label_type).value]
+
     @property
     def multi_label_threshold(self):
         return self._multi_label_threshold
@@ -578,32 +680,10 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         else:
             self._multi_label_threshold = {"default": x}
 
-    def forward_loss(self, sentences: Union[List[DT], DT]) -> Tuple[torch.Tensor, int]:
-
-        # make a forward pass to produce embedded data points and labels
-        embedded_data_points, labels = self.forward_pass(sentences)  # type: ignore
-
-        # no loss can be calculated if there are no labels
-        if not any(labels):
-            return torch.tensor(0.0, requires_grad=True, device=flair.device), 1
-
-        # use dropout
-        embedded_data_points = embedded_data_points.unsqueeze(1)
-        embedded_data_points = self.dropout(embedded_data_points)
-        embedded_data_points = self.locked_dropout(embedded_data_points)
-        embedded_data_points = self.word_dropout(embedded_data_points)
-        embedded_data_points = embedded_data_points.squeeze(1)
-
-        # push embedded_data_points through decoder to get the scores
-        scores = self.decoder(embedded_data_points)
-
-        # calculate the loss
-        return self._calculate_loss(scores, labels)
-
-    def _calculate_loss(self, scores, labels) -> Tuple[torch.Tensor, int]:
-
+    def _prepare_label_tensor(self, prediction_data_points: List[DT2]) -> torch.Tensor:
+        labels = [self._get_label_of_datapoint(dp) for dp in prediction_data_points]
         if self.multi_label:
-            labels = torch.tensor(
+            return torch.tensor(
                 [
                     [1 if label in all_labels_for_point else 0 for label in self.label_dictionary.get_items()]
                     for all_labels_for_point in labels
@@ -611,9 +691,8 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
                 dtype=torch.float,
                 device=flair.device,
             )
-
         else:
-            labels = torch.tensor(
+            return torch.tensor(
                 [
                     self.label_dictionary.get_idx_for_item(label[0])
                     if len(label) > 0
@@ -624,10 +703,56 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
                 device=flair.device,
             )
 
-        return self.loss_function(scores, labels), len(labels)
+    def _encode_data_points(self, sentences: List[DT], data_points: List[DT2]):
+        # embed sentences
+        if self.should_embed_sentence:
+            self.embeddings.embed(sentences)
+
+        # get a tensor of data points
+        data_point_tensor = torch.stack([self._get_embedding_for_data_point(data_point) for data_point in data_points])
+
+        # do dropout
+        data_point_tensor = data_point_tensor.unsqueeze(1)
+        data_point_tensor = self.dropout(data_point_tensor)
+        data_point_tensor = self.locked_dropout(data_point_tensor)
+        data_point_tensor = self.word_dropout(data_point_tensor)
+        data_point_tensor = data_point_tensor.squeeze(1)
+
+        return data_point_tensor
+
+    def _mask_scores(self, scores, data_points):
+        return scores
+
+    def forward_loss(self, sentences: List[DT]) -> Tuple[torch.Tensor, int]:
+        # make a forward pass to produce embedded data points and labels
+        sentences = [sentence for sentence in sentences if self._filter_data_point(sentence)]
+
+        # get the data points for which to predict labels
+        data_points = self._get_data_points_for_batch(sentences)
+        if len(data_points) == 0:
+            return torch.tensor(0.0, requires_grad=True, device=flair.device), 1
+
+        # get their gold labels as a tensor
+        label_tensor = self._prepare_label_tensor(data_points)
+        if label_tensor.size(0) == 0:
+            return torch.tensor(0.0, requires_grad=True, device=flair.device), 1
+
+        # pass data points through network to get encoded data point tensor
+        data_point_tensor = self._encode_data_points(sentences, data_points)
+
+        # decode
+        scores = self.decoder(data_point_tensor)
+
+        # an optional masking step (no masking in most cases)
+        scores = self._mask_scores(scores, data_points)
+
+        # calculate the loss
+        return self._calculate_loss(scores, label_tensor)
+
+    def _calculate_loss(self, scores: torch.Tensor, labels: torch.Tensor) -> Tuple[torch.Tensor, int]:
+        return self.loss_function(scores, labels), labels.size(0)
 
     def _sort_data(self, data_points: List[DT]) -> List[DT]:
-
         if len(data_points) == 0:
             return []
 
@@ -638,7 +763,7 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         sentences = [sentence for sentence in typing.cast(List[Sentence], data_points) if len(sentence) > 0]
 
         # reverse sort all sequences by their length
-        reordered_sentences = sorted(sentences, key=lambda s: len(s), reverse=True)
+        reordered_sentences = sorted(sentences, key=len, reverse=True)
 
         return typing.cast(List[DT], reordered_sentences)
 
@@ -673,12 +798,15 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
             if not isinstance(sentences, list):
                 sentences = [sentences]
 
+            if isinstance(sentences[0], Sentence):
+                Sentence.set_context_for_sentences(typing.cast(List[Sentence], sentences))
+
             reordered_sentences = self._sort_data(sentences)
 
             if len(reordered_sentences) == 0:
                 return sentences
 
-            if len(sentences) > mini_batch_size:
+            if len(reordered_sentences) > mini_batch_size:
                 batches: Union[DataLoader, List[List[DT]]] = DataLoader(
                     dataset=FlairDatapointDataset(reordered_sentences),
                     batch_size=mini_batch_size,
@@ -694,22 +822,31 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
             overall_loss = torch.zeros(1, device=flair.device)
             label_count = 0
             for batch in batches:
+                # filter data points in batch
+                batch = [dp for dp in batch if self._filter_data_point(dp)]
+
                 # stop if all sentences are empty
                 if not batch:
                     continue
 
-                embedded_data_points, gold_labels, data_points = self.forward_pass(  # type: ignore
-                    batch, for_prediction=True
-                )
+                data_points = self._get_data_points_for_batch(batch)
+
+                if not data_points:
+                    continue
+
+                # pass data points through network and decode
+                data_point_tensor = self._encode_data_points(batch, data_points)
+                scores = self.decoder(data_point_tensor)
+                scores = self._mask_scores(scores, data_points)
+
                 # if anything could possibly be predicted
                 if len(data_points) > 0:
-                    scores = self.decoder(embedded_data_points)
-
                     # remove previously predicted labels of this type
-                    for data_point in data_points:
-                        data_point.remove_labels(label_name)
+                    for sentence in data_points:
+                        sentence.remove_labels(label_name)
 
                     if return_loss:
+                        gold_labels = self._prepare_label_tensor(data_points)
                         overall_loss += self._calculate_loss(scores, gold_labels)[0]
                         label_count += len(data_points)
 
@@ -775,6 +912,8 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
             "multi_label",
             "multi_label_threshold",
             "loss_weights",
+            "train_on_gold_pairs_only",
+            "inverse_model",
         ]:
             if arg not in kwargs and arg in state:
                 kwargs[arg] = state[arg]
@@ -791,7 +930,26 @@ class DefaultClassifier(Classifier[DT], typing.Generic[DT]):
         state["multi_label"] = self.multi_label
         state["multi_label_threshold"] = self.multi_label_threshold
         state["loss_weights"] = self.loss_weights
+        state["train_on_gold_pairs_only"] = self.train_on_gold_pairs_only
+        state["inverse_model"] = self.inverse_model
         if self._custom_decoder:
             state["decoder"] = self.decoder
 
         return state
+
+    @classmethod
+    def load(cls, model_path: Union[str, Path, Dict[str, Any]]) -> "DefaultClassifier":
+        from typing import cast
+
+        return cast("DefaultClassifier", super().load(model_path=model_path))
+
+
+def get_non_abstract_subclasses(cls):
+    all_subclasses = []
+    for subclass in cls.__subclasses__():
+        all_subclasses.extend(get_non_abstract_subclasses(subclass))
+        if inspect.isabstract(subclass):
+            continue
+        all_subclasses.append(subclass)
+
+    return all_subclasses
