@@ -1,6 +1,9 @@
 from pathlib import Path
 
 import pytest
+import torch
+
+import flair
 
 
 @pytest.fixture(scope="module")
@@ -28,6 +31,12 @@ def results_base_path(resources_path):
             path.rmdir()
 
 
+@pytest.fixture(autouse=True)
+def set_cpu(force_cpu):
+    if force_cpu:
+        flair.device = torch.device("cpu")
+
+
 def pytest_addoption(parser):
     parser.addoption(
         "--runintegration",
@@ -35,6 +44,12 @@ def pytest_addoption(parser):
         default=False,
         help="run integration tests",
     )
+    parser.addoption(
+        "--force-cpu",
+        action="store_true",
+        default=False,
+        help="use cpu for tests even when gpu is available",
+    )
 
 
 def pytest_collection_modifyitems(config, items):
@@ -43,3 +58,9 @@ def pytest_collection_modifyitems(config, items):
         for item in items:
             if "integration" in item.keywords:
                 item.add_marker(skip_integration)
+
+
+def pytest_generate_tests(metafunc):
+    option_value = metafunc.config.getoption("--force-cpu")
+    if "force_cpu" in metafunc.fixturenames and option_value is not None:
+        metafunc.parametrize("force_cpu", [option_value])
